# vfscript/training/preparer.py
from __future__ import annotations
from dataclasses import dataclass, field
from pathlib import Path
from typing import Callable, Dict, Iterable, List, Optional, Tuple
import json
import math
import numpy as np

# ==== Opcionales ====
try:
    from scipy.spatial import KDTree, ConvexHull  # type: ignore
    _HAS_SCIPY = True
except Exception:
    _HAS_SCIPY = False
    KDTree = None
    ConvexHull = None

try:
    from scipy.special import sph_harm  # type: ignore
    _HAS_SPH = True
except Exception:
    _HAS_SPH = False
    sph_harm = None


# ---------------- Config dataclasses ----------------
@dataclass
class CoordinationCfg:
    enabled: bool = True
    rc: float = 3.25  # Å

@dataclass
class EnergyCfg:
    enabled: bool = True
    # mapeo de nombres posibles de columna de energía en tu dump
    column_candidates: List[str] = field(default_factory=lambda: ["pe", "c_1", "epot", "energy", "c_energy"])

@dataclass
class SteinhardtCfg:
    enabled: bool = True
    radius: float = 2.70  # Å
    orders: Dict[str, bool] = field(default_factory=lambda: {"Q4": True, "Q6": True, "Q8": False, "Q10": False, "Q12": False})

@dataclass
class ConvexHullCfg:
    enabled: bool = False
    area: bool = True
    volume: bool = True

@dataclass
class FeaturesCfg:
    coordination: CoordinationCfg = field(default_factory=CoordinationCfg)
    energy_potential: EnergyCfg = field(default_factory=EnergyCfg)
    steinhardt: SteinhardtCfg = field(default_factory=SteinhardtCfg)
    convex_hull: ConvexHullCfg = field(default_factory=ConvexHullCfg)

@dataclass
class PerfectNetCfg:
    lattice: str = "fcc"     # "fcc" | "bcc"
    a0: float = 3.52         # Å
    cells: List[int] = field(default_factory=lambda: [1, 1, 1])
    atom: str = "Fe"

@dataclass
class TrainingSetup:
    iterations: int = 1000
    max_vacancies: int = 0
    features: FeaturesCfg = field(default_factory=FeaturesCfg)
    perfect_network: PerfectNetCfg = field(default_factory=PerfectNetCfg)

    @staticmethod
    def from_dict(d: Dict) -> "TrainingSetup":
        f = d.get("features", {})
        st = f.get("steinhardt", {})
        ch = f.get("convex_hull", {})
        co = f.get("coordination", {})
        ep = f.get("energy_potential", {})

        return TrainingSetup(
            iterations=int(d.get("iterations", 1000)),
            max_vacancies=int(d.get("max_vacancies", 0)),
            features=FeaturesCfg(
                coordination=CoordinationCfg(**{
                    "enabled": bool(co.get("enabled", True)),
                    "rc": float(co.get("rc", 3.25)),
                }),
                energy_potential=EnergyCfg(**{
                    "enabled": bool(ep.get("enabled", True)),
                    # podés inyectar candidatos desde tu GUI si querés
                }),
                steinhardt=SteinhardtCfg(**{
                    "enabled": bool(st.get("enabled", True)),
                    "radius": float(st.get("radius", 2.70)),
                    "orders": dict(st.get("orders", {"Q4": True, "Q6": True})),
                }),
                convex_hull=ConvexHullCfg(**{
                    "enabled": bool(ch.get("enabled", False)),
                    "area": bool(ch.get("area", True)),
                    "volume": bool(ch.get("volume", True)),
                }),
            ),
            perfect_network=PerfectNetCfg(**{
                "lattice": str(d.get("perfect_network", {}).get("lattice", "fcc")),
                "a0": float(d.get("perfect_network", {}).get("a0", 3.52)),
                "cells": list(d.get("perfect_network", {}).get("cells", [1,1,1])),
                "atom": str(d.get("perfect_network", {}).get("atom", "Fe")),
            })
        )


# ---------------- Utilidades de dump muy mínimas ----------------
class MinimalDumpIO:
    """
    Lector/escritor de .dump (formato ITEM:) con soporte básico:
    - coords (x,y,z)
    - box ortogonal [0,Lx]x[0,Ly]x[0,Lz]
    - columnas extra opcionales (dict de arrays)
    """
    @staticmethod
    def write_simple_dump(path: Path, coords: np.ndarray, box: Tuple[float,float,float],
                          extra_cols: Optional[Dict[str, np.ndarray]] = None) -> None:
        path = Path(path); path.parent.mkdir(parents=True, exist_ok=True)
        n = coords.shape[0]
        Lx, Ly, Lz = box
        cols = ["id", "type", "x", "y", "z"]
        extra_cols = extra_cols or {}
        cols += list(extra_cols.keys())

        with path.open("w", encoding="utf-8") as f:
            f.write("ITEM: TIMESTEP\n0\n")
            f.write("ITEM: NUMBER OF ATOMS\n%d\n" % n)
            f.write("ITEM: BOX BOUNDS pp pp pp\n")
            f.write(f"0 {Lx}\n0 {Ly}\n0 {Lz}\n")
            f.write("ITEM: ATOMS " + " ".join(cols) + "\n")
            for i in range(n):
                row = [str(i+1), "1",
                       f"{coords[i,0]:.8f}", f"{coords[i,1]:.8f}", f"{coords[i,2]:.8f}"]
                for k in extra_cols:
                    row.append(str(extra_cols[k][i]))
                f.write(" ".join(row) + "\n")

    @staticmethod
    def read_coords_and_table(path: Path) -> Tuple[np.ndarray, Dict[str, np.ndarray], Tuple[float,float,float]]:
        """
        Devuelve (coords, tabla, (Lx,Ly,Lz)).
        tabla: dict nombre_columna -> ndarray alineado con coords.
        """
        path = Path(path)
        coords = []
        table_cols = []
        table_vals: List[List[float]] = []
        Lx = Ly = Lz = None  # type: ignore
        with path.open("r", encoding="utf-8", errors="ignore") as f:
            line = f.readline()
            while line:
                if line.startswith("ITEM: BOX BOUNDS"):
                    Lx = float(f.readline().split()[1]);  # segunda col es el max
                    Ly = float(f.readline().split()[1])
                    Lz = float(f.readline().split()[1])
                if line.startswith("ITEM: ATOMS"):
                    parts = line.strip().split()
                    table_cols = parts[2:]  # después de "ITEM: ATOMS"
                    # esperamos filas
                    for row in f:
                        if row.startswith("ITEM:"):
                            break
                        vals = row.strip().split()
                        # mapeo por nombres
                        id_idx = table_cols.index("id") if "id" in table_cols else None
                        x_idx = table_cols.index("x")
                        y_idx = table_cols.index("y")
                        z_idx = table_cols.index("z")
                        coords.append([float(vals[x_idx]), float(vals[y_idx]), float(vals[z_idx])])
                        row_vals = []
                        for j, name in enumerate(table_cols):
                            try:
                                row_vals.append(float(vals[j]))
                            except Exception:
                                row_vals.append(np.nan)
                        table_vals.append(row_vals)
                    break
                line = f.readline()
        coords = np.array(coords, dtype=float)
        table = {}
        if table_cols and table_vals:
            arr = np.array(table_vals, dtype=float)
            for j, name in enumerate(table_cols):
                table[name] = arr[:, j]
        box = (Lx or 0.0, Ly or 0.0, Lz or 0.0)
        return coords, table, box


# ---------------- Clase principal ----------------
class TrainingPreparer:
    def __init__(self,
                 setup: TrainingSetup,
                 out_dir: Path,
                 logger: Optional[Callable[[str], None]] = None):
        self.setup = setup
        self.out_dir = Path(out_dir)
        self.logger = logger or (lambda m: None)
        self.out_dir.mkdir(parents=True, exist_ok=True)
        (self.out_dir / "cache").mkdir(exist_ok=True)

    # ---------- API de construcción ----------
    @staticmethod
    def from_setup_dict(d: Dict, out_dir: Path, logger: Optional[Callable[[str], None]] = None) -> "TrainingPreparer":
        return TrainingPreparer(TrainingSetup.from_dict(d), out_dir, logger)

    def validate(self) -> None:
        s = self.setup
        assert s.iterations > 0, "iterations debe ser > 0"
        assert all(isinstance(c, int) and c >= 1 for c in s.perfect_network.cells), "cells deben ser enteros >=1"
        assert s.perfect_network.lattice.lower() in ("fcc", "bcc"), "lattice debe ser fcc o bcc"

    def prepare_workspace(self) -> None:
        meta = self.out_dir / "meta_training_setup.json"
        with meta.open("w", encoding="utf-8") as f:
            json.dump(self._to_serializable(), f, indent=2)
        self.log(f"Workspace listo. Metadatos en: {meta}")

    # ---------- Red perfecta ----------
    def generate_perfect_dump(self, out_path: Optional[Path] = None) -> Path:
        p = self.setup.perfect_network
        a0 = float(p.a0)
        rx, ry, rz = map(int, p.cells)
        pts = self._make_lattice_points(p.lattice, a0, rx, ry, rz)
        Lx, Ly, Lz = rx * a0, ry * a0, rz * a0
        out_path = Path(out_path or (self.out_dir / "relax_structure.dump"))
        MinimalDumpIO.write_simple_dump(out_path, pts, (Lx, Ly, Lz))
        self.log(f"Dump perfecto generado: {out_path}  (átomos: {pts.shape[0]})")
        return out_path

    # ---------- Dataset / features ----------
    def extract_features_from_dump(self, dump_path: Path) -> Dict[str, float]:
        coords, table, box = MinimalDumpIO.read_coords_and_table(Path(dump_path))
        feats: Dict[str, float] = {}

        # Coordinación (promedio global, como ejemplo)
        if self.setup.features.coordination.enabled:
            rc = self.setup.features.coordination.rc
            neigh_counts = self._coordination_counts(coords, rc)
            feats["coord_mean_rc=%.2f" % rc] = float(np.nanmean(neigh_counts)) if len(neigh_counts) else np.nan

        # Energía potencial (promedio global)
        if self.setup.features.energy_potential.enabled:
            val = self._pick_energy_column(table, self.setup.features.energy_potential.column_candidates)
            feats["energy_potential_mean"] = float(np.nanmean(val)) if val is not None else np.nan

        # Steinhardt Ql (promedio global por l)
        if self.setup.features.steinhardt.enabled:
            if not (_HAS_SCIPY and _HAS_SPH):
                self.log("Steinhardt omitido (requiere SciPy con scipy.special.sph_harm).")
            else:
                r = self.setup.features.steinhardt.radius
                orders = [int(k[1:]) for k, on in self.setup.features.steinhardt.orders.items() if on]
                q_by_atom = self._steinhardt_all_atoms(coords, r, orders)
                for l in orders:
                    feats[f"Q{l}_mean_r={r:.2f}"] = float(np.nanmean(q_by_atom[l])) if l in q_by_atom else np.nan

        # Casco convexo global
        if self.setup.features.convex_hull.enabled:
            if not _HAS_SCIPY or ConvexHull is None:
                self.log("Convex Hull omitido (requiere scipy.spatial.ConvexHull).")
            else:
                try:
                    hull = ConvexHull(coords)
                    if self.setup.features.convex_hull.area:
                        feats["hull_area"] = float(hull.area)
                    if self.setup.features.convex_hull.volume:
                        feats["hull_volume"] = float(hull.volume)
                except Exception as e:
                    self.log(f"ConvexHull falló: {e}")

        return feats

    def build_dataset_csv(self, dumps: Iterable[Path], out_csv: Optional[Path] = None) -> Path:
        rows: List[Dict[str, float]] = []
        for dp in dumps:
            try:
                feats = self.extract_features_from_dump(dp)
                feats["source"] = str(dp)
                rows.append(feats)
                self.log(f"✔ features: {dp}")
            except Exception as e:
                self.log(f"✖ error en {dp}: {e}")

        # columnas ordenadas
        keys = sorted({k for r in rows for k in r.keys()})
        # ‘source’ al final
        if "source" in keys:
            keys.remove("source")
            keys.append("source")

        import csv
        out_csv = Path(out_csv or (self.out_dir / "training_dataset.csv"))
        out_csv.parent.mkdir(parents=True, exist_ok=True)
        with out_csv.open("w", newline="", encoding="utf-8") as f:
            w = csv.DictWriter(f, fieldnames=keys)
            w.writeheader()
            for r in rows:
                w.writerow({k: r.get(k, np.nan) for k in keys})

        self.log(f"CSV de dataset: {out_csv}")
        return out_csv

    # ---------- Helpers internos ----------
    def _to_serializable(self) -> Dict:
        s = self.setup
        return {
            "iterations": s.iterations,
            "max_vacancies": s.max_vacancies,
            "features": {
                "coordination": {"enabled": s.features.coordination.enabled, "rc": s.features.coordination.rc},
                "energy_potential": {"enabled": s.features.energy_potential.enabled,
                                     "column_candidates": s.features.energy_potential.column_candidates},
                "steinhardt": {"enabled": s.features.steinhardt.enabled,
                               "radius": s.features.steinhardt.radius,
                               "orders": s.features.steinhardt.orders},
                "convex_hull": {"enabled": s.features.convex_hull.enabled,
                                "area": s.features.convex_hull.area,
                                "volume": s.features.convex_hull.volume},
            },
            "perfect_network": {
                "lattice": s.perfect_network.lattice,
                "a0": s.perfect_network.a0,
                "cells": s.perfect_network.cells,
                "atom": s.perfect_network.atom,
            }
        }

    def _make_lattice_points(self, lattice: str, a0: float, rx: int, ry: int, rz: int) -> np.ndarray:
        lattice = (lattice or "fcc").lower()
        if lattice not in ("fcc", "bcc"):
            lattice = "fcc"
        if lattice == "fcc":
            basis = np.array([[0,0,0],[0,0.5,0.5],[0.5,0,0.5],[0.5,0.5,0]], float)
        else:
            basis = np.array([[0,0,0],[0.5,0.5,0.5]], float)
        ii, jj, kk = np.mgrid[0:rx, 0:ry, 0:rz]
        cells = np.stack([ii.ravel(), jj.ravel(), kk.ravel()], axis=1).astype(float)
        pos = (cells[:,None,:] + basis[None,:,:]).reshape(-1,3) * a0
        return pos

    def _coordination_counts(self, coords: np.ndarray, rc: float) -> np.ndarray:
        n = coords.shape[0]
        if n == 0:
            return np.array([], dtype=float)

        if _HAS_SCIPY and KDTree is not None:
            tree = KDTree(coords)
            # count neighbors within rc (excluding self)
            counts = np.array([max(0, len(tree.query_ball_point(coords[i], rc)) - 1) for i in range(n)], dtype=float)
            return counts

        # Fallback O(N^2)
        counts = np.zeros(n, dtype=float)
        for i in range(n):
            diffs = coords - coords[i]
            d2 = np.einsum("ij,ij->i", diffs, diffs)
            c = float(np.count_nonzero((d2 > 0.0) & (d2 <= rc*rc)))
            counts[i] = c
        return counts

    def _pick_energy_column(self, table: Dict[str, np.ndarray], candidates: List[str]) -> Optional[np.ndarray]:
        if not table:
            return None
        lowmap = {k.lower(): k for k in table.keys()}
        for name in candidates:
            if name.lower() in lowmap:
                return table[lowmap[name.lower()]]
        # heurística: primera col que contenga 'e' y sea numérica
        for k in table:
            if "e" in k.lower():
                v = table[k]
                if isinstance(v, np.ndarray) and np.issubdtype(v.dtype, np.number):
                    return v
        return None

    def _steinhardt_all_atoms(self, coords: np.ndarray, r: float, orders: List[int]) -> Dict[int, np.ndarray]:
        """
        Implementación estándar:
        q_l(i) = sqrt((4π/(2l+1)) * sum_m |q_lm(i)|^2 ), con
        q_lm(i) = (1/N_b(i)) sum_{j in nb(i)} Y_lm(theta_ij, phi_ij).
        """
        if not (_HAS_SCIPY and KDTree is not None and sph_harm is not None):
            raise RuntimeError("Steinhardt requiere SciPy (KDTree + sph_harm).")

        n = coords.shape[0]
        tree = KDTree(coords)
        # vecinos por átomo
        nbs = [tree.query_ball_point(coords[i], r) for i in range(n)]
        # eliminamos self
        for i in range(n):
            if i in nbs[i]:
                nbs[i].remove(i)

        # preparar salida
        q: Dict[int, np.ndarray] = {l: np.zeros(n, dtype=float) for l in orders}

        # cálculo por átomo
        for i in range(n):
            nbrs = nbs[i]
            Nb = len(nbrs)
            if Nb == 0:
                for l in orders:
                    q[l][i] = np.nan
                continue
            r_ij = coords[nbrs] - coords[i]
            # esféricas
            r_norm = np.linalg.norm(r_ij, axis=1)
            # evitar divisiones por cero
            mask = r_norm > 0
            if not np.any(mask):
                for l in orders:
                    q[l][i] = np.nan
                continue
            x, y, z = r_ij[mask, 0], r_ij[mask, 1], r_ij[mask, 2]
            theta = np.arccos(np.clip(z / np.linalg.norm(r_ij[mask], axis=1), -1.0, 1.0))  # [0,pi]
            phi = np.mod(np.arctan2(y, x), 2*np.pi)  # [0,2pi)

            for l in orders:
                qlm_sum = 0+0j
                # promediamos sobre m y vecinos:
                qlm_sq_sum = 0.0
                for m in range(-l, l+1):
                    Y = sph_harm(m, l, phi, theta)  # complejo
                    qlm = np.mean(Y)  # promedio sobre vecinos
                    qlm_sq_sum += (np.abs(qlm)**2).real
                q[i] = q.get(l, q[l])
                q[l][i] = math.sqrt((4*math.pi/(2*l+1)) * qlm_sq_sum)
        return q

    def log(self, msg: str) -> None:
        self.logger(str(msg))



import json, pandas as pd
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split, RandomizedSearchCV, cross_val_score
from sklearn.metrics import classification_report
import joblib

class ImprovedVacancyClassifier:
    def __init__(self, json_path, random_state=42):
        self.json_path = json_path
        self.random_state = random_state
        self.features = ['surface_area', 'filled_volume', 'cluster_size']
        self.label_map = {0:"1-3",1:"4-6",2:"7-9",3:"10+"}

        # Pipeline base
        self.pipe = Pipeline([
            ('scaler', StandardScaler()),
            ('clf', RandomForestClassifier(class_weight='balanced', random_state=self.random_state))
        ])

        # Distribución de búsqueda
        self.param_dist = {
            'clf__n_estimators': [50,100,200],
            'clf__max_depth': [None,10,20],
            'clf__min_samples_split': [2,5,10]
        }

    def load_and_prepare(self):
        data = json.load(open(self.json_path))
        df = pd.DataFrame(data)
        df['y'] = df['vacancys'].apply(self._clasificar)
        X = df[self.features]
        y = df['y']
        return X, y

    def _clasificar(self, vac):
        if 1<=vac<=3: return 0
        if 4<=vac<=6: return 1
        if 7<=vac<=9: return 2
        return 3

    def train(self):
        X, y = self.load_and_prepare()
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.3, random_state=self.random_state
        )

        search = RandomizedSearchCV(
            self.pipe, self.param_dist,
            n_iter=20, cv=5, scoring='f1_macro',
            n_jobs=-1, random_state=self.random_state
        )
        search.fit(X_train, y_train)
        self.model = search.best_estimator_

        # Evaluar en test
        y_pred = self.model.predict(X_test)
        # print("==> Mejoros parámetros:")
        # print(search.best_params_)
        # print("\n==> Reporte en test:")
        # print(classification_report(y_test, y_pred, target_names=self.label_map.values()))

        # Guardar
        joblib.dump(self.model, "best_vacancy_model.pkl")

    def predict(self, X_new):
        numeric = self.model.predict(X_new)
        return [self.label_map[i] for i in numeric]
    def classify_csv(self, csv_path, output_path=None):
        """
        Lee un CSV con las columnas de features, predice el grupo y (opcionalmente)
        vuelca un nuevo CSV con la columna 'grupo_predicho'.
        """
        df = pd.read_csv(csv_path)
        missing = set(self.features) - set(df.columns)
        if missing:
            raise ValueError(f"Faltan columnas en el CSV: {missing}")

        X_new = df[self.features]
        pred_numeric = self.model.predict(X_new)
        df['grupo_predicho'] = [self.label_map[i] for i in pred_numeric]

        if output_path:
            df.to_csv(output_path, index=False)
            print(f"Archivo guardado en {output_path}")

        return df



# modifiers/training/vacancy_predictors.py

import os
import json
import math
import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import train_test_split, KFold, cross_val_score
from sklearn.preprocessing import StandardScaler
from sklearn.pipeline import Pipeline
import xgboost as xgb
from vfscript.training.utils import load_json_data, resolve_input_params_path


class VacancyPredictorRF:
    def __init__(
        self,
        json_path: str = "outputs.vfinder/training_data.json",
        predictor_columns: list = None
    ):
        """
        Predictor de vacancias con RandomForest. Usa las columnas que se definen
        en predictor_columns (se obtienen de input_params.json).
        """
        self.json_path = json_path

        
        if predictor_columns is None:
            
            raise ValueError("Debes pasar predictor_columns o definirlas explícitamente.")
        self.columns = predictor_columns

        
        self.df = load_json_data(self.json_path)

        
        self.model = self._train_model()


    def _train_model(self):
        X = self.df[self.columns]
        y = self.df["vacancys"]
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.3, random_state=42
        )
        model = RandomForestRegressor(random_state=42)
        model.fit(X_train, y_train)
        y_pred = model.predict(X_test)
        mse = mean_squared_error(y_test, y_pred)
        return model


    @staticmethod
    def _round_up(x):
        return math.ceil(x)


    def predict_vacancies(self, **kwargs):
        data = {col: [kwargs[col]] for col in self.columns}
        nuevos_datos = pd.DataFrame(data)
        prediction = self.model.predict(nuevos_datos)[0]
        #print(f"Predicción inicial: {prediction}")
        return self._round_up(prediction)




class XGBoostVacancyPredictor:
    def __init__(
        self,
        training_data_path: str = "outputs.vfinder/training_data.json",
        model_path: str = "outputs.json/xgboost_model.json",
        predictor_columns: list = None,
        n_splits: int = 5,
        random_state: int = 42
    ):
        """
        Predictor de vacancias con XGBoost. Hace cross‐validation y guarda el modelo
        en 'model_path'.
        """
        self.training_data_path = training_data_path
        self.model_path = model_path
        self.n_splits = n_splits
        self.random_state = random_state

        if predictor_columns is None:
            raise ValueError("Debes pasar predictor_columns o definirlas explícitamente.")
        self.columns = predictor_columns

        self.scaler = StandardScaler()
        self.model = xgb.XGBRegressor(
            objective='reg:squarederror',
            random_state=self.random_state,
            n_estimators=100,
            learning_rate=0.1,
            max_depth=4,
            subsample=0.8,
            colsample_bytree=0.8
        )
        self._load_data_and_train()


    def _load_data_and_train(self):
        with open(self.training_data_path, "r", encoding="utf-8") as f:
            data = json.load(f)

        feature_list = []
        for col in self.columns:
            if col in data:
                feature_list.append(data[col])
            else:
                raise ValueError(f"No existe la columna '{col}' en los datos de entrenamiento.")

        X = np.column_stack(feature_list)
        y = np.array(data["vacancys"])

        X = self.scaler.fit_transform(X)
        
        n_samples = X.shape[0]
        n_splits = self.n_splits if n_samples >= self.n_splits else max(n_samples, 2)
        kfold = KFold(n_splits=n_splits, shuffle=True, random_state=self.random_state)

        scores = cross_val_score(self.model, X, y, scoring='neg_mean_squared_error', cv=kfold)
        mse_scores = -scores  
        

        X_train, X_val, y_train, y_val = train_test_split(
            X, y, test_size=0.2, random_state=self.random_state
        )
        self.model.fit(X_train, y_train, eval_set=[(X_val, y_val)], verbose=False)
        
        os.makedirs(os.path.dirname(self.model_path), exist_ok=True)
        self.model.save_model(self.model_path)


    def predict(self, sample_input: np.ndarray) -> np.ndarray:
        sample_input = np.array(sample_input)
        sample_input = self.scaler.transform(sample_input)
        prediction = self.model.predict(sample_input)
        #print(f"Predicción inicial: {prediction}")
        return prediction




class VacancyPredictor:
    def __init__(self, json_path: str = "outputs.vfinder/training_data.json"):
        """
        Predictor de vacancias con regresión lineal simple, usando solo 'surface_area'.
        """
        self.json_path = json_path
        self.columns = ["surface_area"]
        self.df = load_json_data(self.json_path)
        self.model = self._train_model()


    def _train_model(self):
        X = self.df[self.columns]
        y = self.df["vacancys"]
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.3, random_state=42
        )
        model = LinearRegression()
        model.fit(X_train, y_train)
        y_pred = model.predict(X_test)
        mse = mean_squared_error(y_test, y_pred)
        
        return model


    @staticmethod
    def _round_positive(x):
        return math.ceil(x) if x > 0 else math.ceil(-x)


    def predict_vacancies(self, **kwargs):
        nuevos_datos = pd.DataFrame({col: [kwargs[col]] for col in self.columns})
        prediction = self.model.predict(nuevos_datos)[0]
        #print(f"Predicción inicial: {prediction}")
        return self._round_positive(prediction)




from sklearn.neural_network import MLPRegressor

class VacancyPredictorMLP:
    def __init__(
        self,
        json_path: str = "outputs.vfinder/training_data.json",
        predictor_columns: list = None
    ):
        """
        Predictor de vacancias con red neuronal (MLP). Crea un pipeline que escala
        y entrena un MLPRegressor.
        """
        self.json_path = json_path

        if predictor_columns is None:
            raise ValueError("Debes pasar predictor_columns o definirlas explícitamente.")
        self.columns = predictor_columns

        self.df = load_json_data(self.json_path)
        self.model = self._train_model()


    def _train_model(self):
        X = self.df[self.columns]
        y = self.df["vacancys"]

        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=0.3, random_state=42
        )

        pipeline = Pipeline([
            ('scaler', StandardScaler()),
            ('mlp', MLPRegressor(
                hidden_layer_sizes=(128, 64),
                activation='relu',
                solver='adam',
                learning_rate_init=0.01,
                max_iter=1000,
                early_stopping=True,
                n_iter_no_change=20,
                random_state=42
            ))
        ])
        pipeline.fit(X_train, y_train)
        y_pred = pipeline.predict(X_test)
        mse = mean_squared_error(y_test, y_pred)
        #print("MSE del modelo MLP:", mse)

        return pipeline


    @staticmethod
    def _round_up(x):
        return math.ceil(x) if x > 0 else math.ceil(-x)


    def predict_vacancies(self, **kwargs):
        data = pd.DataFrame({col: [kwargs[col]] for col in self.columns})
        prediction = self.model.predict(data)[0]
        #print(f"Predicción inicial: {prediction}")  
        return self._round_up(prediction)



# main.py

from .core import *  
from .config_loader import cargar_json_usuario
from pathlib import Path
import os
import pandas as pd
import warnings
import argparse
warnings.filterwarnings('ignore', message='.*OVITO.*PyPI')

def VacancyAnalysis():
    
    base = "outputs"
    for sub in ("csv", "dump", "json"):
        os.makedirs(os.path.join(base, sub), exist_ok=True)

    


    
    CONFIG = cargar_json_usuario()
    
    if "CONFIG" not in CONFIG or not isinstance(CONFIG["CONFIG"], list) or len(CONFIG["CONFIG"]) == 0:
        raise ValueError("input_params.json debe contener una lista 'CONFIG' con al menos un objeto.")

    configuracion = CONFIG["CONFIG"][0]
    raw_defects = configuracion.get('defect', [])  
    defect_files = raw_defects if isinstance(raw_defects, list) else [raw_defects]  
    for FILE in defect_files:
        cs_out_dir = Path("inputs")
        cs_generator = CrystalStructureGenerator(configuracion, cs_out_dir)
        dump_path = cs_generator.generate()
        #print(f"Estructura relajada generada en: {dump_path}")
        if configuracion['training']:
        # 1) Recuperar el dict que guarda tu pestaña (CONFIG[0]['training_setup'])
     
            params = json.load(open("input_params.json", "r", encoding="utf-8"))
            setup_dict = params["CONFIG"][0]["training_setup"]

            # 2) Construir el preparador
            prep = TrainingPreparer.from_setup_dict(setup_dict, out_dir=Path("outputs/training"),
                                                    logger=lambda m: print("[training]", m))

            # 3) Validar + preparar entorno
            prep.validate()
            prep.prepare_workspace()

            # 4) (Opcional) generar una red perfecta .dump para pruebas
            dump_ref = prep.generate_perfect_dump()  # outputs/training/relax_structure.dump

            # 5) Extraer features de uno o varios dumps y generar CSV
            csv_path = prep.build_dataset_csv([dump_ref])   # o bien una lista de dumps reales tuyos
            print("Dataset listo en:", csv_path)
            gen = AtomicGraphGenerator()
            gen.run()
            p = argparse.ArgumentParser(description="Exporta features (normas + ConvexHull) a CSV.")
            p.add_argument("--output-csv", default="outputs/csv/finger_data.csv",
                        help="Ruta del CSV de salida.")
            p.add_argument("--dump", action="append",
                        help="Ruta a uno o más .dump (repetible). Si se omite, se toma de input_params.json.")
            args = p.parse_args()

            exporter = FeatureExporter(
                dump_paths=args.dump if args.dump else None,
                output_csv=args.output_csv
            )
            try:
                exporter.export()
            except ImportError as e:
                # Si falta SciPy, el compute_convex_hull lo avisa acá
                print(f"[ERROR] {e}\nSugerencia: pip install scipy")


        analyzer = DeformationAnalyzer(FILE, configuracion['generate_relax'][0], configuracion['generate_relax'][5], threshold=0.02)
        delta = analyzer.compute_metric()
        method = analyzer.select_method()
        #print(f"Métrica δ = {delta:.4f}, método seleccionado: {method}")
        






        # 2) Condicional
        if method == 'geometric' and configuracion['geometric_method']:
            # Aplico el método geométrico
            vac_analyzer = WSMet(
                defect_dump_path=FILE,
                lattice_type=configuracion['generate_relax'][0],
                element=configuracion['generate_relax'][5],
                tolerance=0.5
            )
            vacancies = vac_analyzer.run()
            # vacancies es la lista de posiciones
        elif method == 'ml' or configuracion['geometric_method']==False :
            vac_analyzer = WSMet(
                defect_dump_path=FILE,
                lattice_type=configuracion['generate_relax'][0],
                element=configuracion['generate_relax'][5],
                tolerance=0.5
            )
            vac_analyzer.generate_perfect_atoms()
            
            processor = ClusterProcessor(FILE)
            processor.run()
            separator = KeyFilesSeparator(configuracion, os.path.join("outputs/json", "clusters.json"))
            separator.run()

            # 3. Procesar dumps críticos
            clave_criticos = ClusterDumpProcessor.cargar_lista_archivos_criticos("outputs/json/key_archivos.json")
            for archivo in clave_criticos:
                try:
                    dump_proc = ClusterDumpProcessor(archivo, decimals=5)
                    dump_proc.load_data()
                    dump_proc.process_clusters()
                    dump_proc.export_updated_file(f"{archivo}_actualizado.txt")
                except Exception as e:
                    print(f"Error procesando {archivo}: {e}")

            # 4. Subdivisión iterativa
            lista_criticos = ClusterDumpProcessor.cargar_lista_archivos_criticos("outputs/json/key_archivos.json")
            for archivo in lista_criticos:
                machine_proc = ClusterProcessorMachine(archivo)  # o ClusterProcessorMachine(archivo, "input_params.json")
                machine_proc.process_clusters()
                machine_proc.export_updated_file()


            # 5. Separar archivos finales vs críticos
            separator = KeyFilesSeparator(configuracion, os.path.join("outputs/json", "clusters.json"))
            separator.run()

            # 6. Generar nuevos dumps por cluster
            export_list = ExportClusterList("outputs/json/key_archivos.json")
            export_list.process_files()

            # 7. Calcular superficies de dump
            surf_proc = SurfaceProcessor(configuracion)
            surf_proc.process_all_files()
            surf_proc.export_results()


            exporter = ClusterFeatureExporter("outputs/json/key_archivos.json")
            exporter.export()


            # ------------------------------------------------------------------------
            # 8. Entrenar y clasificar defectos con el nuevo modelo ensemble
            if configuracion['geometric_method']:
                analyzer = WSMet("inputs/void_15.dump", "bcc", "Fe", tolerance=0.5)
                vac_positions = analyzer.run()
            # Instancia y entrena el modelo
            clf = ImprovedVacancyClassifier(json_path='outputs/json/training_graph.json')
            clf.train()  
            # → ya imprime mejores parámetros y reporte de clasificación

            # Clasifica tu CSV de defectos (añade columna 'grupo_predicho')
            df_clasif = clf.classify_csv(
                csv_path='outputs/csv/defect_data.csv',
                output_path='outputs/csv/finger_data_clasificado.csv'
            )
            #print(df_clasif[['archivo','grupo_predicho']])
            # ------------------------------------------------------------------------
            # 9. Predicción de vacancias según grupo_predicho
            # Como ImprovedVacancyClassifier usa la misma API de train/classify,
            # podemos reusar el método classify_csv para predecir vacancias
            # (o bien, si lo prefieres, crear un método predict_from_csv
            #  similar al anterior)
            df_pred = clf.classify_csv(
                csv_path='outputs/csv/finger_data_clasificado.csv',
                output_path='outputs/csv/finger_data_predicha.csv'
            )
            #print("✅ Vacancias predichas guardadas en outputs/csv/finger_data_predicha.csv")

            # ------------------------------------------------------------------------
           
            assigner = FingerprintVacancyAssigner(
                base_csv_path="outputs/csv/finger_data.csv",
                query_csv_path="outputs/csv/finger_key_files.csv",
                weight_N=1
            )
            df_result = assigner.assign()
            df_result.to_csv("outputs/csv/finger_key_files_clasificado.csv", index=False)
            #print("✅ Resultado guardado con peso_N =", assigner.weight_N)
            #Calcular categoria de defect_file.csv
            #model = BehaviorTreeModel(weight_cluster_size=2.0, max_depth=5)
            #model.train('outputs/json/training_graph.json')

            #df_resultado = model.classify_csv(
            #   csv_path='outputs/csv/defect_data.csv',
            #  output_path='outputs/csv/finger_data_clasificado.csv'
            #)

            #print(df_resultado[['archivo', 'grupo_predicho']])


            #ETAPA DE PREDICCIONES

            trainer = VacancyModelTrainer(json_path="outputs/json/training_graph.json")
            trainer.load_data()
            trainer.train_group_classifier()
            trainer.train_all_regressors()
        
            # Predicción por CSV
            out_df = trainer.predict_from_csv("outputs/csv/finger_data_clasificado.csv")

            out_df.to_csv("outputs/csv/results.csv", index=False)
            #print("✅ Guardado results.csv")
        # Ejemplo de uso
            assigner = FingerprintVacancyAssigner(
                base_csv_path="outputs/csv/finger_data.csv",
                query_csv_path="outputs/csv/finger_key_files.csv",
                weight_N=100 # peso mayor para 'N'
            )
            df_result = assigner.assign()
            df_result.to_csv("outputs/csv/finger_key_files_clasificado.csv", index=False)
            #print("✅ Resultado guardado con peso_N =", assigner.weight_N)

        else:
            raise RuntimeError(f"Método desconocido: {method}")
        

        #ESTIMACION POR CLASIGICACION DE COEFICIENTE DE VACANCIA AREA
        calc = GroupCoefficientCalculator("outputs/json/training_graph.json")
        calc.load()
        # Usa mínimos teóricos (1,4,7,10) y redondeo hacia arriba (ceil)
        out = calc.estimate_from_defect_csv(
            defect_csv_path="outputs/csv/finger_data_clasificado.csv",
            group_col=None,                 # detecta entre ['grupo_predicho','grupo','Group','label','group']
            surface_area_col="surface_area",
            out_path="outputs/csv/defect_data_estimated.csv",
            use_observed_min_instead=False, # pon True si querés dividir por el min OBSERVADO del grupo
            round_mode="ceil"
        )
        #print(out.head())

            

if __name__ == "__main__":
    VacancyAnalysis()
    print("Script ejecutado correctamente.")






from ase.io.lammpsrun import read_lammps_dump
from scipy.spatial import cKDTree
import numpy as np

class DeformationAnalyzer:
    """
    Clase para calcular la métrica de deformación δ de una estructura atómica LAMMPS dump
    y decidir automáticamente si usar el método geométrico o el modelo de ML.
    """

    def __init__(self, dump_path: str, lattice_type: str, element: str, threshold: float = 0.02):
        """
        Parámetros:
        - dump_path: ruta al archivo LAMMPS dump defectuoso.
        - lattice_type: 'bcc' o 'fcc'.
        - element: símbolo del elemento (ej. 'Fe').
        - threshold: umbral δ para definir deformación alta.
        """
        self.dump_path = dump_path
        self.lattice_type = lattice_type.lower()
        self.element = element
        self.threshold = threshold

        # Cargar átomos
        self.atoms = read_lammps_dump(self.dump_path, index=-1)
        self.cell = self.atoms.get_cell()
        self.volume = self.cell.volume
        self.n_atoms = len(self.atoms)

        # Calcular parámetro de red y distancia ideal NN
        self.a = self._compute_lattice_parameter()
        self.d_ideal = self._compute_d_ideal()

    def _compute_lattice_parameter(self) -> float:
        """
        Calcula el parámetro de red 'a' estimado desde la densidad atómica.
        """
        natoms_per_cell = 2 if self.lattice_type == 'bcc' else 4
        density = self.n_atoms / self.volume
        unit_cell_volume = natoms_per_cell / density
        return unit_cell_volume ** (1/3)

    def _compute_d_ideal(self) -> float:
        """
        Calcula la distancia ideal al primer vecino para la red definida.
        """
        if self.lattice_type == 'bcc':
            return np.sqrt(3) / 2 * self.a
        elif self.lattice_type == 'fcc':
            return self.a / np.sqrt(2)
        else:
            raise ValueError("Tipo de red no reconocido: usar 'bcc' o 'fcc'.")

    def compute_metric(self) -> float:
        """
        Computa δ = σ_nn / d_ideal, donde σ_nn es la desviación estándar de las
        distancias al vecino más cercano.
        """
        positions = self.atoms.get_positions()
        tree = cKDTree(positions)
        dists, _ = tree.query(positions, k=2)
        nn_dists = dists[:, 1]  # descartamos la distancia cero a sí mismo
        sigma = np.std(nn_dists)
        return sigma / self.d_ideal

    def is_highly_deformed(self) -> bool:
        """
        Devuelve True si δ ≥ threshold, lo que indica muestra "muy deformada".
        """
        return self.compute_metric() >= self.threshold

    def select_method(self) -> str:
        """
        Retorna 'geometric' si la muestra está poco deformada (δ < threshold),
        o 'ml' si está muy deformada.
        """
        return 'ml' if self.is_highly_deformed() else 'geometric'

# Ejemplo de uso:
# analyzer = DeformationAnalyzer("inputs/fe6.dump", "bcc", "Fe", threshold=0.02)
# δ = analyzer.compute_metric()
# method = analyzer.select_method()
# print(f"Métrica δ = {δ:.4f}, método seleccionado: {method}")
